@@ -25,20 +25,26 @@ MAX_RETRIES = 100
 MAGIC_KEY = -1
 
 class SchemaError(Exception):
-    """Raised when a DynamoDBModel class's schema is incorrect."""
-    pass
+    """SchemaError exception is raised when a schema consistency check fails.
+    Most of the checks are performed in :py:meth:`~.ConnectionBorg.create_table`.
 
+    Common consistency failure includes lacks of ``__table__``, ``__hash_key__``,
+    ``__schema__`` definition or when an :py:class:`~.autoincrement_int` ``hash_key``
+    is used with a ``range_key``.
+    """
 
+#FIXME: should be raised when Amazons fails because of this instead of a Boto stuff
 class ThroughputError(Exception):
-    """Raised when reaquested throughput can not be allocated."""
-    pass
+    """Raised when requested throughput can not be allocated by
+    :py:meth:`~.ConnectionBorg.create_table`. It probably means that either read
+    or write is below Amazon's minimum of 5.
+    """
 
 
 class MaxRetriesExceededError(Exception):
     """Raised when a failed operation couldn't be completed after retrying
-    MAX_RETRIES times (e.g. saving an autoincrementing hash_key).
+    ``MAX_RETRIES`` times (e.g. saving an autoincrementing hash_key).
     """
-    pass
 
 
 class OverwriteError(Exception):
@@ -46,7 +52,6 @@ class OverwriteError(Exception):
     in the database and we've forbidden that because we believe we're creating
     a new one (see :meth:`DynamoDBModel.save`).
     """
-    pass
 
 
 class ExpectedValueError(Exception):
@@ -55,7 +60,6 @@ class ExpectedValueError(Exception):
     doesn't match what is stored in the database (i.e. when somebody changed
     the DB's version of your object behind your back).
     """
-    pass
 
 
 class autoincrement_int(int):
@@ -71,19 +75,19 @@ class autoincrement_int(int):
 
     Auto-incrementing int keys are implemented by storing a special "magic"
     item in the table with the following properties:
-      - hash_key_value = -1
-      - __max_hash_key__ = X
-    where X is the maximum used hash_key value.
+
+        - ``hash_key_value = -1``
+        - ``__max_hash_key__ = N``
+
+    where N is the maximum used hash_key value.
 
     Inserting a new item issues an atomic add on the '__max_hash_key__' value.
     Its new value is returned and used as the primary key for the new elem.
 
-    Note that hash_key_value is set to '-1' while '__max_hash_key__' initial
+    Note that hash_key_value is set to '-1' while ``__max_hash_key__`` initial
     value is 0. This will element at key '0' unused. It's actually a garbage item
     for cases where a value is manually added to an unitialized index.
     """
-    pass
-
 
 _JSON_TYPES = frozenset([list, dict])
 
@@ -130,12 +134,13 @@ def _get_default_value(schema_type, default=None):
       - For strings, it's an empty string.
       - For numbers, it's zero.
 
-      This function may raise TypeError exception if:
+    This function may raise TypeError exception if:
+
        - default was callable and required arguments
        - default or its return value is not an instance of schema_type
 
-      :param schema_type class object to instanciate
-      :param default default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
+    :param schema_type class object to instanciate
+    :param default default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
 
     """
     if default is not None:
@@ -209,7 +214,9 @@ def _dynamodb_to_python(schema_type, value):
 
     ``_dynamodb_to_python(t, _python_to_dynamodb(v)) == v`` for any v.
 
-    :param schema_type: A type supported by the mapper (TODO Clearly list those).
+    :param schema_type: A type supported by the mapper
+
+    .. (TODO Clearly list those).
 
     :param value: The DynamoDB attribute to convert to a Python object.
         May be ``None``.
@@ -241,12 +248,13 @@ def _dynamodb_to_python(schema_type, value):
 class ConnectionBorg(object):
     """Borg that handles access to DynamoDB.
 
-    You should never make any explicit/direct boto.dynamodb calls by yourself
+    You should never make any explicit/direct ``boto.dynamodb`` calls by yourself
     except for table maintenance operations :
-    * boto.dynamodb.table.update_throughput()
-    * boto.dynamodb.table.delete()
 
-    Remember to call :meth:`set_auth_credentials`, or to set the
+        - ``boto.dynamodb.table.update_throughput()``
+        - ``boto.dynamodb.table.delete()``
+
+    Remember to call :meth:`set_credentials`, or to set the
     ``AWS_ACCESS_KEY_ID`` and ``AWS_SECRET_ACCESS_KEY`` environment variables
     before making any calls.
     """
@@ -284,15 +292,24 @@ class ConnectionBorg(object):
         item.put({item.hash_key_name: False})
 
     def set_credentials(self, aws_access_key_id, aws_secret_access_key):
-        """Set the DynamoDB credentials."""
+        """Set the DynamoDB credentials. If boto is already configured on this
+        machine, this step is optional.
+        Access keys can be found in `Amazon's console.
+        <https://aws-portal.amazon.com/gp/aws/developer/account/index.html?action=access-key>`_
+
+        :param aws_access_key_id: AWS api access key ID
+
+        :param aws_secret_access_key: AWS api access key
+
+        """
         self._aws_access_key_id = aws_access_key_id
         self._aws_secret_access_key = aws_secret_access_key
 
     def create_table(self, cls, read_units, write_units, wait_for_active=False):
         """Create a table that'll be used to store instances of cls.
 
-        See `http://docs.amazonwebservices.com/amazondynamodb/latest/developerguide/ProvisionedThroughputIntro.html`_
-        for information about provisioned throughput.
+        See `Amazon's developer guide <http://docs.amazonwebservices.com/amazondynamodb/latest/developerguide/ProvisionedThroughputIntro.html>`_
+        for more information about provisioned throughput.
 
         :param cls: The class whose instances will be stored in the table.
 
@@ -314,6 +331,7 @@ class ConnectionBorg(object):
         if not table_name:
             raise SchemaError("Class does not define __table__", cls)
 
+        # FIXME: check key is defined in schema
         if not hash_key_name:
             raise SchemaError("Class does not define __hash_key__", cls)
 
@@ -360,6 +378,8 @@ class ConnectionBorg(object):
         if hash_key_type is autoincrement_int:
             self._create_autoincrement_magic_item(table)
 
+        dblog.debug("Created table %s(%s, %s)", cls.__table__, hash_key_name, range_key_name)
+
         return table
 
     def get_table(self, name):
@@ -377,15 +397,15 @@ class DynamoDBModel(object):
 
     Each subclass must define the following attributes:
 
-      - __table__: the name of the table used for storage.
-      - __hash_key__: the name of the primary hash key.
-      - __range_key__: (optional) if you're using a composite primary key,
+      - ``__table__``: the name of the table used for storage.
+      - ``__hash_key__``: the name of the primary hash key.
+      - ``__range_key__``: (optional) if you're using a composite primary key,
           the name of the range key.
-      - __schema__: {attribute_name: attribute_type} mapping.
+      - ``__schema__``: ``{attribute_name: attribute_type}`` mapping.
           Supported attribute_types are: int, long, float, str, unicode, set.
           Default values are obtained by calling the type with no args
           (so 0 for numbers, "" for strings and empty sets).
-      - __defaults__: {attribute_name: defaulter} optional mapping.
+      - ``__defaults__``: (optional) ``{attribute_name: defaulter}`` mapping.
           This dict allows to provide a default value for each attribute_name at
           object creation time. It will *never* be used when loading from the DB.
           It is fully optional. If no value is supplied the empty value
@@ -402,14 +422,41 @@ class DynamoDBModel(object):
     we're storing empty sets/strings as missing attributes in DynamoDB, and
     converting back and forth based on the schema.
 
-    So if your schema looks like the following: {"id": unicode, "name" str,
-    "cheats": set}, then {"id": "e1m1", "name": "Hangar",
-    "cheats": set(["idkfa", "iddqd"])} will be stored exactly as is, but
-    {"id": "e1m2", "name": "", "cheats": set()} will be stored as simply
-    {"id": "e1m2"}
+    So if your schema looks like the following::
+
+        {
+            "id": unicode,
+            "name": str,
+            "cheats": set
+        }
+
+    then::
+
+        {
+            "id": "e1m1",
+            "name": "Hangar",
+            "cheats": set([
+                "idkfa",
+                "iddqd"
+            ])
+        }
+
+    will be stored exactly as is, but::
+
+        {
+            "id": "e1m2",
+            "name": "",
+            "cheats": set()
+        }
+
+    will be stored as simply::
+
+        {
+            "id": "e1m2"
+        }
 
 
-    TODO Add checks for common error cases:
+    .. TODO Add checks for common error cases:
         - Wrong datatypes in the schema
         - hash_key/range_key incorrectly defined
     """
@@ -426,9 +473,9 @@ class DynamoDBModel(object):
         """Create an instance of the model. All fields defined in the schema
         are created. By order of prioritym its value will be loaded from:
 
-        * kwargs
-        * __default__
-        * mapper's default (0, empty string, empty set, ...)
+            - kwargs
+            - __defaults__
+            - mapper's default (0, empty string, empty set, ...)
 
         We're supplying this method to avoid the need for extra checks in save and
         ease object initial creation.
@@ -451,6 +498,7 @@ class DynamoDBModel(object):
         Default values are used for anything that's missing from the dict
         (see DynamoDBModel class docstring).
         """
+        #FIXME: type check. moving to __init__ syntax may break some impementations
         #FIXME: rename to from_db_dict when breaking compat or disambiguing
         instance = cls()
         for (name, type_) in cls.__schema__.iteritems():
@@ -491,7 +539,7 @@ class DynamoDBModel(object):
                     range_key=r_value,
                     consistent_read=consistent_read)
 
-        dblog.debug("[GET] %s(%s, %s)", cls.__table__, h_value, r_value)
+        dblog.debug("Got item (%s, %s) from table %s", h_value, r_value, cls.__table__)
 
         return cls.from_dict(item)
 
@@ -503,14 +551,17 @@ class DynamoDBModel(object):
         primary key(s) for each object you want to retrieve:
 
           - If the primary keys are hash keys, keys must be a list of
-            their values (e.g. [1, 2, 3, 4]).
+            their values (e.g. ``[1, 2, 3, 4]``).
           - If the primary keys are composite (hash + range), keys must
-            be a list of (hash_key, range_key) values
-            (e.g. [("user1", 1), ("user1", 2), ("user1", 3)]).
+            be a list of ``(hash_key, range_key)`` values
+            (e.g. ``[("user1", 1), ("user1", 2), ("user1", 3)]``).
 
         get_batch *always* performs eventually consistent reads.
 
         Please not that a batch can *not* read more than 100 items at once.
+
+        :param keys: iterable of keys. ex ``[(hash1, range1), (hash2, range2)]``
+
         """
         if len(keys) > 100:
             raise ValueError("Too many items to read in a single batch. Maximum is 100.")
@@ -533,14 +584,14 @@ class DynamoDBModel(object):
 
         res = batch_list.submit()
 
-        dblog.debug("[BATCH GET] %s", cls.__table__)
+        dblog.debug("Sent a batch get on table %s", cls.__table__)
 
         return [
             cls.from_dict(d) for d in res[u"Responses"][cls.__table__][u"Items"]
         ]
 
     @classmethod
-    def query(cls, hash_key_value, range_key_condition=None, consistent_read=False, reverse=False):
+    def query(cls, hash_key_value, range_key_condition=None, consistent_read=False, reverse=False, limit=None):
         """Query DynamoDB for items matching the requested key criteria.
 
         You need to supply an exact hash key value, and optionally, conditions
@@ -555,12 +606,28 @@ class DynamoDBModel(object):
         :param hash_key_value: The hash key's value for all requested items.
 
         :param range_key_condition: A condition instance from
-            boto.dynamodb.condition -- one of EQ(x), LE(x), LT(x), GE(x),
-            GT(x), BEGINS_WITH(x), BETWEEN(x, y).
+            ``boto.dynamodb.condition`` -- one of
+
+                - EQ(x)
+                - LE(x)
+                - LT(x)
+                - GE(x)
+                - GT(x)
+                - BEGINS_WITH(x)
+                - BETWEEN(x, y)
 
         :param consistent_read: If False (default), an eventually consistent
             read is performed. Set to True for strongly consistent reads.
 
+        :param reverse: Ask DynamoDB to scan the ``range_key`` in the reverse
+            order. For example, if you use dates here, the more recent element
+            will be returned first. Defaults to ``False``.
+
+        :param limit: Specify the maximum number of items to read from the table.
+            Even though Boto returns a generator, it works by batchs of 1MB.
+            using this option may help to spare some read credits. Defaults to
+            ``None``
+
         :rtype: generator
         """
         table = ConnectionBorg().get_table(cls.__table__)
@@ -570,9 +637,10 @@ class DynamoDBModel(object):
                 h_value,
                 range_key_condition,
                 consistent_read=consistent_read,
-                scan_index_forward=not reverse)
+                scan_index_forward=not reverse,
+                max_results=limit)
 
-        dblog.debug("[QUERY] %s(%s, %s)", cls.__table__, h_value, range_key_condition)
+        dblog.debug("Queried (%s, %s) on table %s", h_value, range_key_condition, cls.__table__)
 
         return (cls.from_dict(d) for d in res)
 
@@ -586,8 +654,8 @@ class DynamoDBModel(object):
         Scan is a very expensive operation -- it doesn't use any indexes and will
         look through the entire table. As much as possible, you should avoid it.
 
-        :param scan_filter: A {attribute_name: condition} dict, where
-            condition is a condition instance from boto.dynamodb.condition.
+        :param scan_filter: A ``{attribute_name: condition}`` dict, where
+            condition is a condition instance from ``boto.dynamodb.condition``.
 
         :rtype: generator
         """
@@ -596,7 +664,7 @@ class DynamoDBModel(object):
 
         res = table.scan(scan_filter)
 
-        dblog.debug("[QUERY] %s(%s)", cls.__table__, scan_filter)
+        dblog.debug("Scanned table %s with filter %s", cls.__table__, scan_filter)
 
         return (
             cls.from_dict(d)
@@ -644,8 +712,10 @@ class DynamoDBModel(object):
     def _save_autoincrement_hash_key(self, item):
         """Compute an autoincremented hash_key for an item and save it to the DB.
 
-        To achieve this goal, we keep a special object at hash_key=0 to keep track
-        of the counter status. We then issue an atomic inc to the counter field.
+        To achieve this goal, we keep a special object at ``hash_key=MAGIC_KEY``
+        to keep track of the counter status. We then issue an atomic inc to the
+        counter field.
+
         We do not need to read it befor as we know its hesh_key yet.
         The new value is send back to us and used as the hash_key for elem
         """
@@ -681,7 +751,7 @@ class DynamoDBModel(object):
         """Save the object to the database.
 
         This method may be used both to insert a new object in the DB, or to
-        update an existing one (iff allow_overwrite == True).
+        update an existing one (iff ``allow_overwrite == True``).
 
         :param allow_overwrite: If False, the method will only succeed if this
             object's primary keys don't exist in the database (otherwise,
@@ -706,7 +776,7 @@ class DynamoDBModel(object):
 
         # Detect magic elem manual overwrite
         if schema[hash_key] == autoincrement_int and item_data[hash_key] == MAGIC_KEY:
-            raise SchemaError()
+            raise SchemaError()#FIXME: probably not the best exception to throw
         # We're inserting a new item in an autoincrementing table.
         if schema[hash_key] == autoincrement_int and item_data[hash_key] == 0:
             # Compute the index and save the object
@@ -743,6 +813,10 @@ class DynamoDBModel(object):
                 # Unhandled exception
                 raise
 
+        hash_key_value = getattr(self, hash_key)
+        range_key_value = getattr(self, hash_key, None)
+        dblog.debug("Saved (%s, %s) in table %s allow_overwrite=%s", hash_key_value, range_key_value, cls.__table__, allow_overwrite)
+
     def delete(self):
         """Delete the current object from the database."""
         cls = type(self)
@@ -755,3 +829,5 @@ class DynamoDBModel(object):
 
         table = ConnectionBorg().get_table(cls.__table__)
         Item(table, hash_key_value, range_key_value).delete()
+
+        dblog.debug("Deleted (%s, %s) from table %s", hash_key_value, range_key_value, cls.__table__)
